from datetime import datetime
import json
from typing import Any, Callable
from unittest.mock import ANY, MagicMock

import pytest

import pykoplenti


def test_me_parsing():
    raw_response = """\
    {
        "role": "NONE",
        "anonymous": true,
        "locked": false,
        "permissions": [],
        "active": false,
        "authenticated": false
    }"""

    me = pykoplenti.MeData(**json.loads(raw_response))

    assert me.is_locked is False
    assert me.is_active is False
    assert me.is_authenticated is False
    assert me.permissions == []
    assert me.is_anonymous is True
    assert me.role == "NONE"


def test_version_parsing():
    raw_response = """\
    {
        "sw_version": "01.26.09454",
        "name": "PUCK RESTful API",
        "api_version": "0.2.0",
        "hostname": "scb"
    }"""

    version = pykoplenti.VersionData(**json.loads(raw_response))

    assert version.api_version == "0.2.0"
    assert version.hostname == "scb"
    assert version.name == "PUCK RESTful API"
    assert version.sw_version == "01.26.09454"


def test_event_parsing():
    raw_response = """\
    {
        "description": "Reduction of AC power due to external command.",
        "category": "info",
        "is_active": false,
        "code": 5014,
        "end_time": "2023-04-29T00:45:19",
        "start_time": "2023-04-29T00:44:18",
        "group": "Information",
        "long_description": "Reduction of AC power due to external command."
    }"""

    event = pykoplenti.EventData(**json.loads(raw_response))

    assert event.start_time == datetime(2023, 4, 29, 0, 44, 18)
    assert event.end_time == datetime(2023, 4, 29, 0, 45, 19)
    assert event.is_active is False
    assert event.code == 5014
    assert event.long_description == "Reduction of AC power due to external command."
    assert event.category == "info"
    assert event.description == "Reduction of AC power due to external command."
    assert event.group == "Information"


def test_module_parsing():
    raw_response = """\
    {
        "id": "devices:local:powermeter",
        "type": "device:powermeter"
    }"""

    module = pykoplenti.ModuleData(**json.loads(raw_response))

    assert module.id == "devices:local:powermeter"
    assert module.type == "device:powermeter"


def test_process_parsing():
    raw_response = """\
    {
        "id": "Inverter:State",
        "unit": "",
        "value": 6
    }"""

    process_data = pykoplenti.ProcessData(**json.loads(raw_response))

    assert process_data.id == "Inverter:State"
    assert process_data.unit == ""
    assert process_data.value == 6


def test_settings_parsing():
    raw_response = """\
    {
        "min": "0",
        "default": null,
        "access": "readonly",
        "unit": null,
        "id": "Properties:PowerId",
        "type": "uint32",
        "max": "100000"
    }"""

    settings_data = pykoplenti.SettingsData(**json.loads(raw_response))

    assert settings_data.unit is None
    assert settings_data.default is None
    assert settings_data.id == "Properties:PowerId"
    assert settings_data.max == "100000"
    assert settings_data.min == "0"
    assert settings_data.type == "uint32"
    assert settings_data.access == "readonly"


def test_process_data_list():
    json = [
        {"id": "Statistic:Yield:Day", "unit": "%", "value": 1},
        {"id": "Statistic:Yield:Month", "unit": "%", "value": 2},
    ]

    assert pykoplenti.model.process_data_list(json) == [
        pykoplenti.ProcessData(id="Statistic:Yield:Day", unit="%", value="1"),
        pykoplenti.ProcessData(id="Statistic:Yield:Month", unit="%", value="2"),
    ]


def test_process_data_collection_indicates_length():
    raw_response = (
        '[{"id": "Statistic:Yield:Day", "unit": "", "value": 1}, '
        '{"id": "Statistic:Yield:Month", "unit": "", "value": 2}]'
    )
    pdc = pykoplenti.ProcessDataCollection(
        pykoplenti.model.process_data_list(json.loads(raw_response))
    )

    assert len(pdc) == 2


def test_process_data_collection_index_returns_processdata():
    raw_response = (
        '[{"id": "Statistic:Yield:Day", "unit": "", "value": 1}, '
        '{"id": "Statistic:Yield:Month", "unit": "", "value": 2}]'
    )
    pdc = pykoplenti.ProcessDataCollection(
        pykoplenti.model.process_data_list(json.loads(raw_response))
    )

    result = pdc["Statistic:Yield:Month"]

    assert isinstance(result, pykoplenti.ProcessData)
    assert result.id == "Statistic:Yield:Month"
    assert result.unit == ""
    assert result.value == 2


def test_process_data_collection_can_be_iterated():
    raw_response = (
        '[{"id": "Statistic:Yield:Day", "unit": "", "value": 1}, '
        '{"id": "Statistic:Yield:Month", "unit": "", "value": 2}]'
    )
    pdc = pykoplenti.ProcessDataCollection(
        pykoplenti.model.process_data_list(json.loads(raw_response))
    )

    result = list(pdc)

    assert result == ["Statistic:Yield:Day", "Statistic:Yield:Month"]


@pytest.mark.asyncio
async def test_relogin_on_401_response(
    pykoplenti_client: MagicMock,
    client_response_factory: Callable[[int, Any], MagicMock],
):
    """Ensures that a re-login is executed if a 401 response was returned."""

    # First response returns 401
    client_response_factory(401, None)

    # Second response is successfull
    client_response_factory(
        200,
        [
            {
                "moduleid": "moda",
                "processdata": [{"id": "procb", "unit": "", "value": 0}],
            }
        ],
    )

    _ = await pykoplenti_client.get_process_data_values("moda", "procb")

    pykoplenti_client._login.assert_awaited_once()


@pytest.mark.asyncio
async def test_process_data_value(
    pykoplenti_client: MagicMock,
    client_response_factory: Callable[[int, Any], MagicMock],
    websession: MagicMock,
):
    """Test if process data values could be retrieved."""
    client_response_factory(
        200,
        [
            {
                "moduleid": "devices:local:pv1",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 700.0},
                ],
            },
            {
                "moduleid": "devices:local:pv2",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 300.0},
                ],
            },
        ],
    )

    values = await pykoplenti_client.get_process_data_values(
        {"devices:local:pv1": ["P"], "devices:local:pv2": ["P"]}
    )

    websession.request.assert_called_once_with(
        "POST",
        ANY,
        headers=ANY,
        json=[
            {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
            {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
        ],
    )

    assert values == {
        "devices:local:pv1": pykoplenti.ProcessDataCollection(
            [pykoplenti.ProcessData(id="P", unit="W", value=700.0)]
        ),
        "devices:local:pv2": pykoplenti.ProcessDataCollection(
            [pykoplenti.ProcessData(id="P", unit="W", value=300.0)]
        ),
    }



from base64 import b64decode, b64encode
from collections.abc import Mapping
import contextlib
from datetime import datetime
import functools
import hashlib
import hmac
import locale
import logging
from os import urandom
from typing import IO, Dict, Final, Iterable, List, Union, overload
import warnings

from Crypto.Cipher import AES
from aiohttp import ClientResponse, ClientSession, ClientTimeout
from yarl import URL

from .model import (
    EventData,
    MeData,
    ModuleData,
    ProcessDataCollection,
    SettingsData,
    VersionData,
    process_data_list,
)

_logger: Final = logging.getLogger(__name__)


class ApiException(Exception):
    """Base exception for API calls."""

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return f"API Error: {self.msg}"


class InternalCommunicationException(ApiException):
    """Exception for internal communication error response."""

    def __init__(self, status_code: int, error: str):
        super().__init__(f"Internal communication error ([{status_code}] - {error})")
        self.status_code = status_code
        self.error = error


class AuthenticationException(ApiException):
    """Exception for authentication or user error response."""

    def __init__(self, status_code: int, error: str):
        super().__init__(
            f"Invalid user/Authentication failed ([{status_code}] - {error})"
        )
        self.status_code = status_code
        self.error = error


class NotAuthorizedException(ApiException):
    """Exception for calles without authentication."""

    def __init__(self, status_code: int, error: str):
        super().__init__(f"Not authorized ([{status_code}] - {error})")
        self.status_code = status_code
        self.error = error


class UserLockedException(ApiException):
    """Exception for user locked error response."""

    def __init__(self, status_code: int, error: str):
        super().__init__(f"User is locked ([{status_code}] - {error})")
        self.status_code = status_code
        self.error = error


class ModuleNotFoundException(ApiException):
    """Exception for module or setting not found response."""

    def __init__(self, status_code: int, error: str):
        super().__init__(f"Module or setting not found ([{status_code}] - {error})")
        self.status_code = status_code
        self.error = error


def _relogin(fn):
    """Decorator for automatic re-login if session was expired."""

    @functools.wraps(fn)
    async def _wrapper(self: "ApiClient", *args, **kwargs):
        with contextlib.suppress(AuthenticationException, NotAuthorizedException):
            return await fn(self, *args, **kwargs)
        _logger.debug("Request failed - try to re-login")
        await self._login()
        return await fn(self, *args, **kwargs)

    return _wrapper


class ApiClient(contextlib.AbstractAsyncContextManager):
    """Client for the REST-API of Kostal Plenticore inverters.

    The RESP-API provides several scopes of information. Each scope provides a
    dynamic set of data which can be retrieved using this interface. The scopes
    are:

    - process data (readonly, dynamic values of the operation)
    - settings (some are writable, static values for configuration)

    The data are grouped into modules. For example the module `devices:local`
    provides a process data `Dc_P` which contains the value of the current
    DC power.

    To get all process data or settings the methods `get_process_data` or
    `get_settings` can be used. Depending of the current logged in user the
    returned data can vary.

    The methods `get_process_data_values` and `get_setting_values` can be used
    to read process data or setting values from the inverter. You can use
    `set_setting_values` to write new setting values to the inverter if the
    setting is writable.

    The authorization system of the inverter comprises three states:
    * not logged in (is_active=False, authenticated=False)
    * logged in and active (is_active=True, authenticated=True)
    * logged in and inactive (is_active=False, authenticated=False)

    The current state can be queried with the `get_me` method. Depending of
    this state some operation might not be available.
    """

    BASE_URL = "/api/v1/"
    SUPPORTED_LANGUAGES = {
        "de": ["de"],
        "en": ["gb"],
        "es": ["es"],
        "fr": ["fr"],
        "hu": ["hu"],
        "it": ["it"],
        "nl": ["nl"],
        "pl": ["pl"],
        "pt": ["pt"],
        "cs": ["cz"],
        "el": ["gr"],
        "zh": ["cn"],
    }

    def __init__(self, websession: ClientSession, host: str, port: int = 80):
        """Create a new client.

        :param websession: A aiohttp ClientSession for all requests
        :param host: The hostname or ip of the inverter
        :param port: The port of the API interface (default 80)
        """
        self.websession = websession
        self.host = host
        self.port = port
        self.session_id: Union[str, None] = None
        self._key: Union[str, None] = None
        self._service_code: Union[str, None] = None
        self._user: Union[str, None] = None

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Logout support for context manager."""
        if self.session_id is not None:
            await self.logout()

    def _create_url(self, path: str) -> URL:
        """Creates a REST-API URL with the given path as suffix.

        :param path: path suffix, must not start with '/'
        :return: a URL instance
        """
        base = URL.build(
            scheme="http",
            host=self.host,
            port=self.port,
            path=ApiClient.BASE_URL,
        )
        return base.join(URL(path))

    async def initialize_virtual_process_data(self):
        process_data = await self.get_process_data()
        self._virt_process_data.initialize(process_data)

    async def login(
        self,
        key: str,
        service_code: Union[str, None] = None,
        password: Union[str, None] = None,
        user: Union[str, None] = None,
    ):
        """Login with the given password (key).

        If a service code is provided user is 'master', else 'user'.

        Parameters
        ----------
        :param key: The user password. If 'service_code' is given, 'key' is the
                    Master Key (also called Device ID).
        :type key: str, None
        :param service_code: Installer service code. If given the user is assumed to be
                             'master', else 'user'.
        :type service_code: str, None
        :param password: Deprecated, use key instead.
        :param user: Deprecated, user is chosen automatically depending on service_code.

        :raises AuthenticationException: if authentication failed
        :raises aiohttp.client_exceptions.ClientConnectorError: if host is not reachable
        :raises asyncio.exceptions.TimeoutError: if a timeout occurs
        """

        if password is None:
            self._key = key
        else:
            warnings.warn(
                "password is deprecated. Use key instead.", DeprecationWarning
            )
            self._key = password

        if user is None:
            self._user = "master" if service_code else "user"
        else:
            warnings.warn(
                "user is deprecated. user is chosen automatically.", DeprecationWarning
            )

        self._service_code = service_code

        try:
            await self._login()
        except Exception:
            self._key = None
            self._user = None
            self._service_code = None
            raise

    async def _login(self):
        # Step 1 start authentication
        client_nonce = urandom(12)

        start_request = {
            "username": self._user,
            "nonce": b64encode(client_nonce).decode("utf-8"),
        }

        async with self.websession.request(
            "POST", self._create_url("auth/start"), json=start_request
        ) as resp:
            await self._check_response(resp)
            start_response = await resp.json()
            server_nonce = b64decode(start_response["nonce"])
            transaction_id = b64decode(start_response["transactionId"])
            salt = b64decode(start_response["salt"])
            rounds = start_response["rounds"]

        # Step 2 finish authentication (RFC5802)
        salted_passwd = hashlib.pbkdf2_hmac(
            "sha256", self._key.encode("utf-8"), salt, rounds
        )
        client_key = hmac.new(
            salted_passwd, "Client Key".encode("utf-8"), hashlib.sha256
        ).digest()
        stored_key = hashlib.sha256(client_key).digest()

        auth_msg = (
            "n={user},r={client_nonce},r={server_nonce},s={salt},i={rounds},"
            "c=biws,r={server_nonce}".format(
                user=self._user,
                client_nonce=b64encode(client_nonce).decode("utf-8"),
                server_nonce=b64encode(server_nonce).decode("utf-8"),
                salt=b64encode(salt).decode("utf-8"),
                rounds=rounds,
            )
        )
        client_signature = hmac.new(
            stored_key, auth_msg.encode("utf-8"), hashlib.sha256
        ).digest()
        client_proof = bytes(a ^ b for a, b in zip(client_key, client_signature))

        server_key = hmac.new(
            salted_passwd, "Server Key".encode("utf-8"), hashlib.sha256
        ).digest()
        server_signature = hmac.new(
            server_key, auth_msg.encode("utf-8"), hashlib.sha256
        ).digest()

        finish_request = {
            "transactionId": b64encode(transaction_id).decode("utf-8"),
            "proof": b64encode(client_proof).decode("utf-8"),
        }

        async with self.websession.request(
            "POST", self._create_url("auth/finish"), json=finish_request
        ) as resp:
            await self._check_response(resp)
            finish_response = await resp.json()
            token = finish_response["token"]
            signature = b64decode(finish_response["signature"])
            if signature != server_signature:
                raise Exception("Server signature mismatch.")

        # Step 3 create session
        session_key_hmac = hmac.new(
            stored_key, "Session Key".encode("utf-8"), hashlib.sha256
        )
        session_key_hmac.update(auth_msg.encode("utf-8"))
        session_key_hmac.update(client_key)
        protocol_key = session_key_hmac.digest()
        session_nonce = urandom(16)
        cipher = AES.new(protocol_key, AES.MODE_GCM, nonce=session_nonce)

        if self._user == "master":
            token = f"{token}:{self._service_code}"

        cipher_text, auth_tag = cipher.encrypt_and_digest(token.encode("utf-8"))

        session_request = {
            # AES initialization vector
            "iv": b64encode(session_nonce).decode("utf-8"),
            # AES GCM tag
            "tag": b64encode(auth_tag).decode("utf-8"),
            # ID of authentication transaction
            "transactionId": b64encode(transaction_id).decode("utf-8"),
            # Only the token or token and service code (separated by colon). Encrypted
            # with AES using the protocol key
            "payload": b64encode(cipher_text).decode("utf-8"),
        }

        async with self.websession.request(
            "POST", self._create_url("auth/create_session"), json=session_request
        ) as resp:
            await self._check_response(resp)
            session_response = await resp.json()
            self.session_id = session_response["sessionId"]

    def _session_request(self, path: str, method="GET", **kwargs):
        """Make an request on the current active session.

        :param path: the URL suffix
        :param method: the request method, defaults to 'GET'
        :param **kwargs: all other args are forwarded to the request
        """

        headers: Dict[str, str] = {}
        if self.session_id is not None:
            headers["authorization"] = f"Session {self.session_id}"

        return self.websession.request(
            method, self._create_url(path), headers=headers, **kwargs
        )

    async def _check_response(self, resp: ClientResponse):
        """Check if the given response contains an error and throws
        the appropriate exception."""

        if resp.status == 200:
            return

        try:
            response = await resp.json()
            error = response["message"]
        except Exception:
            error = None

        if resp.status == 400:
            raise AuthenticationException(resp.status, error)

        if resp.status == 401:
            raise NotAuthorizedException(resp.status, error)

        if resp.status == 403:
            raise UserLockedException(resp.status, error)

        if resp.status == 404:
            raise ModuleNotFoundException(resp.status, error)

        if resp.status == 503:
            raise InternalCommunicationException(resp.status, error)

        # we got an undocumented status code
        raise ApiException(f"Unknown API response [{resp.status}] - {error}")

    async def logout(self):
        """Logs the current user out."""
        self._key = None
        self._service_code = None
        async with self._session_request("auth/logout", method="POST") as resp:
            await self._check_response(resp)

    async def get_me(self) -> MeData:
        """Returns information about the user.

        No login is required.
        """
        async with self._session_request("auth/me") as resp:
            await self._check_response(resp)
            me_response = await resp.json()
            return MeData(**me_response)

    async def get_version(self) -> VersionData:
        """Returns information about the API of the inverter.

        No login is required.
        """
        async with self._session_request("info/version") as resp:
            await self._check_response(resp)
            response = await resp.json()
            return VersionData(**response)

    @_relogin
    async def get_events(self, max_count=10, lang=None) -> Iterable[EventData]:
        """Returns a list with the latest localized events.

        :param max_count: the max number of events to read
        :param lang: the RFC1766 based language code, for example 'de_CH' or 'en'
        """
        if lang is None:
            lang = locale.getlocale()[0]

        language = lang[:2].lower()
        variant = lang[3:5].lower()
        if language not in ApiClient.SUPPORTED_LANGUAGES.keys():
            # Fallback to default
            language = "en"
            variant = "gb"
        else:
            variants = ApiClient.SUPPORTED_LANGUAGES[language]
            if variant not in variants:
                variant = variants[0]

        request = {"language": f"{language}-{variant}", "max": max_count}

        async with self._session_request(
            "events/latest", method="POST", json=request
        ) as resp:
            await self._check_response(resp)
            event_response = await resp.json()
            return [EventData(**x) for x in event_response]

    async def get_modules(self) -> Iterable[ModuleData]:
        """Return list of all available modules (providing process data or settings)."""
        async with self._session_request("modules") as resp:
            await self._check_response(resp)
            modules_response = await resp.json()
            return [ModuleData(**x) for x in modules_response]

    @_relogin
    async def get_process_data(self) -> Mapping[str, Iterable[str]]:
        """Return a dictionary of all processdata ids and its module ids.

        :return: a dictionary with the module id as key and a list of process data ids
                 as value
        """
        async with self._session_request("processdata") as resp:
            await self._check_response(resp)
            data_response = await resp.json()
            return {x["moduleid"]: x["processdataids"] for x in data_response}

    @overload
    async def get_process_data_values(
        self,
        module_id: str,
        processdata_id: str,
    ) -> Mapping[str, ProcessDataCollection]: ...

    @overload
    async def get_process_data_values(
        self,
        module_id: str,
        processdata_id: Iterable[str],
    ) -> Mapping[str, ProcessDataCollection]: ...

    @overload
    async def get_process_data_values(
        self,
        module_id: str,
    ) -> Mapping[str, ProcessDataCollection]: ...

    @overload
    async def get_process_data_values(
        self,
        module_id: Mapping[str, Iterable[str]],
    ) -> Mapping[str, ProcessDataCollection]: ...

    @overload
    async def get_process_data_values(
        self,
        module_id: Union[str, Mapping[str, Iterable[str]]],
        processdata_id: Union[str, Iterable[str], None] = None,
    ) -> Mapping[str, ProcessDataCollection]: ...

    @_relogin
    async def get_process_data_values(
        self,
        module_id: Union[str, Mapping[str, Iterable[str]]],
        processdata_id: Union[str, Iterable[str], None] = None,
    ) -> Mapping[str, ProcessDataCollection]:
        """Return a dictionary of process data of one or more modules.

        :param module_id: required, must be a module id or a mapping with the
                          module id as key and the process data ids as values.
        :param processdata_id: optional, if given `module_id` must be string. Can
                               be either a string or a list of string. If missing
                               all process data ids are returned.
        :return: a dictionary with the module id as key and a instance of
                 :py:class:`ProcessDataCollection` as value
        """

        if isinstance(module_id, str) and processdata_id is None:
            # get all process data of a module
            async with self._session_request(f"processdata/{module_id}") as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {
                    data_response[0]["moduleid"]: ProcessDataCollection(
                        process_data_list(data_response[0]["processdata"])
                    )
                }

        if isinstance(module_id, str) and isinstance(processdata_id, str):
            # get a single process data of a module
            async with self._session_request(
                f"processdata/{module_id}/{processdata_id}"
            ) as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {
                    data_response[0]["moduleid"]: ProcessDataCollection(
                        process_data_list(data_response[0]["processdata"])
                    )
                }

        if (
            isinstance(module_id, str)
            and processdata_id is not None
            and hasattr(processdata_id, "__iter__")
        ):
            # get multiple process data of a module
            ids = ",".join(processdata_id)
            async with self._session_request(f"processdata/{module_id}/{ids}") as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {
                    data_response[0]["moduleid"]: ProcessDataCollection(
                        process_data_list(data_response[0]["processdata"])
                    )
                }

        if isinstance(module_id, dict) and processdata_id is None:
            # get multiple process data of multiple modules
            request = []
            for mid, pids in module_id.items():
                # the json encoder expects that iterables are either list or tuples,
                # other types has to be converted
                if isinstance(pids, (list, tuple)):
                    request.append(dict(moduleid=mid, processdataids=pids))
                else:
                    request.append(dict(moduleid=mid, processdataids=list(pids)))

            async with self._session_request(
                "processdata", method="POST", json=request
            ) as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {
                    x["moduleid"]: ProcessDataCollection(
                        process_data_list(x["processdata"])
                    )
                    for x in data_response
                }

        raise TypeError("Invalid combination of module_id and processdata_id.")

    async def get_settings(self) -> Mapping[str, Iterable[SettingsData]]:
        """Return list of all modules with a list of available settings identifiers."""
        async with self._session_request("settings") as resp:
            await self._check_response(resp)
            response = await resp.json()
            result: Dict[str, List[SettingsData]] = {}
            for module in response:
                mid = module["moduleid"]
                data = [SettingsData(**x) for x in module["settings"]]
                result[mid] = data

            return result

    @overload
    async def get_setting_values(
        self,
        module_id: str,
        setting_id: str,
    ) -> Mapping[str, Mapping[str, str]]: ...

    @overload
    async def get_setting_values(
        self,
        module_id: str,
        setting_id: Iterable[str],
    ) -> Mapping[str, Mapping[str, str]]: ...

    @overload
    async def get_setting_values(
        self,
        module_id: str,
    ) -> Mapping[str, Mapping[str, str]]: ...

    @overload
    async def get_setting_values(
        self,
        module_id: Mapping[str, Iterable[str]],
    ) -> Mapping[str, Mapping[str, str]]: ...

    @_relogin
    async def get_setting_values(
        self,
        module_id: Union[str, Mapping[str, Iterable[str]]],
        setting_id: Union[str, Iterable[str], None] = None,
    ) -> Mapping[str, Mapping[str, str]]:
        """Return a dictionary of setting values of one or more modules.

        :param module_id: required, must be a module id or a dictionary with the
                          module id as key and the setting ids as values.
        :param setting_id: optional, if given `module_id` must be string. Can
                           be either a string or a list of string. If missing
                           all setting ids are returned.
        """
        if isinstance(module_id, str) and setting_id is None:
            # get all setting data of a module
            async with self._session_request(f"settings/{module_id}") as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {module_id: {data_response[0]["id"]: data_response[0]["value"]}}

        if isinstance(module_id, str) and isinstance(setting_id, str):
            # get a single setting of a module
            async with self._session_request(
                f"settings/{module_id}/{setting_id}"
            ) as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {module_id: {data_response[0]["id"]: data_response[0]["value"]}}

        if (
            isinstance(module_id, str)
            and setting_id is not None
            and hasattr(setting_id, "__iter__")
        ):
            # get multiple settings of a module
            ids = ",".join(setting_id)
            async with self._session_request(f"settings/{module_id}/{ids}") as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {module_id: {x["id"]: x["value"] for x in data_response}}

        if isinstance(module_id, dict) and setting_id is None:
            # get multiple process data of multiple modules
            request = []
            for mid, pids in module_id.items():
                # the json encoder expects that iterables are either list or tuples,
                # other types has to be converted
                if isinstance(pids, (list, tuple)):
                    request.append(dict(moduleid=mid, settingids=pids))
                else:
                    request.append(dict(moduleid=mid, settingids=list(pids)))

            async with self._session_request(
                "settings", method="POST", json=request
            ) as resp:
                await self._check_response(resp)
                data_response = await resp.json()
                return {
                    x["moduleid"]: {y["id"]: y["value"] for y in x["settings"]}
                    for x in data_response
                }

        raise TypeError("Invalid combination of module_id and setting_id.")

    @_relogin
    async def set_setting_values(self, module_id: str, values: Mapping[str, str]):
        """Write a list of settings for one modules."""
        request = [
            {
                "moduleid": module_id,
                "settings": [dict(value=v, id=k) for k, v in values.items()],
            }
        ]
        async with self._session_request(
            "settings", method="PUT", json=request
        ) as resp:
            await self._check_response(resp)

    @_relogin
    async def download_logdata(
        self,
        writer: IO,
        begin: Union[datetime, None] = None,
        end: Union[datetime, None] = None,
    ):
        """Download logdata as tab-separated file."""
        request = {}
        if begin is not None:
            request["begin"] = begin.strftime("%Y-%m-%d")
        if end is not None:
            request["end"] = end.strftime("%Y-%m-%d")

        async with self._session_request(
            "logdata/download",
            method="POST",
            json=request,
            timeout=ClientTimeout(total=360),
        ) as resp:
            await self._check_response(resp)
            async for data in resp.content.iter_any():
                writer.write(data.decode("UTF-8"))



from .api import (
    ApiClient,
    ApiException,
    AuthenticationException,
    InternalCommunicationException,
    ModuleNotFoundException,
    NotAuthorizedException,
    UserLockedException,
)
from .extended import ExtendedApiClient
from .model import (
    EventData,
    MeData,
    ModuleData,
    ProcessData,
    ProcessDataCollection,
    SettingsData,
    VersionData,
)

__all__ = [
    "MeData",
    "VersionData",
    "ModuleData",
    "ProcessData",
    "ProcessDataCollection",
    "SettingsData",
    "EventData",
    "ApiException",
    "InternalCommunicationException",
    "AuthenticationException",
    "NotAuthorizedException",
    "UserLockedException",
    "ModuleNotFoundException",
    "ApiClient",
    "ExtendedApiClient",
]



"""Smoketest which are executed on a real inverter."""

import os
import re
from typing import AsyncGenerator

import aiohttp
import pytest
import pytest_asyncio

import pykoplenti


@pytest_asyncio.fixture
async def authenticated_client() -> AsyncGenerator[pykoplenti.ApiClient, None]:
    host = os.getenv("SMOKETEST_HOST", "localhost")
    port = int(os.getenv("SMOKETEST_PORT", 80))
    password = os.getenv("SMOKETEST_PASS", "")

    async with aiohttp.ClientSession() as session:
        client = pykoplenti.ExtendedApiClient(session, host, port)
        await client.login(password)
        yield client
        await client.logout()


@pytest.mark.skipif(
    os.getenv("SMOKETEST_HOST") is None, reason="Smoketest must be explicitly executed"
)
class TestSmokeTests:
    """Contains smoke tests which are executed on a real inverter.

    This tests are not automatically executed because they need real HW. Please
    note that all checks are highl volatile because of different configuration and
    firmware version of the inverter.
    """

    @pytest.mark.asyncio
    async def test_smoketest_me(self, authenticated_client: pykoplenti.ApiClient):
        """Retrieves the MeData."""

        me = await authenticated_client.get_me()

        assert me == pykoplenti.MeData(
            locked=False,
            active=True,
            authenticated=True,
            permissions=[],
            anonymous=False,
            role="USER",
        )

    @pytest.mark.asyncio
    async def test_smoketest_version(self, authenticated_client: pykoplenti.ApiClient):
        """Retrieves the VersionData."""

        version = await authenticated_client.get_version()

        # version info are highly variable hence only some basic checks are performed
        assert len(version.hostname) > 0
        assert len(version.name) > 0
        assert re.match(r"\d+.\d+.\d+", version.api_version) is not None
        assert re.match(r"\d+.\d+.\d+", version.sw_version) is not None

    @pytest.mark.asyncio
    async def test_smoketest_modules(self, authenticated_client: pykoplenti.ApiClient):
        """Retrieves the ModuleData."""

        modules = list(await authenticated_client.get_modules())

        assert len(modules) >= 17
        assert pykoplenti.ModuleData(id="devices:local", type="device") in modules

    @pytest.mark.asyncio
    async def test_smoketest_settings(self, authenticated_client: pykoplenti.ApiClient):
        """Retrieves the SettingsData."""

        settings = await authenticated_client.get_settings()

        assert "devices:local" in settings
        assert (
            pykoplenti.SettingsData(
                min="0",
                max="32",
                default=None,
                access="readonly",
                unit=None,
                id="Branding:ProductName1",
                type="string",
            )
            in settings["devices:local"]
        )

    @pytest.mark.asyncio
    async def test_smoketest_setting_value1(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Retrieves the setting value with variante 1."""

        setting_value = await authenticated_client.get_setting_values(
            "devices:local", "Branding:ProductName1"
        )

        assert setting_value == {
            "devices:local": {"Branding:ProductName1": "PLENTICORE plus"}
        }

    @pytest.mark.asyncio
    async def test_smoketest_setting_value2(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Retrieves the setting value with variante 2."""

        setting_value = await authenticated_client.get_setting_values(
            "devices:local", ["Branding:ProductName1"]
        )

        assert setting_value == {
            "devices:local": {"Branding:ProductName1": "PLENTICORE plus"}
        }

    @pytest.mark.asyncio
    @pytest.mark.skip(reason="API endpoint is not working")
    async def test_smoketest_setting_value3(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Retrieves the setting value with variante 3."""

        setting_value = await authenticated_client.get_setting_values(
            "devices:local"
        )

        assert (
            setting_value["devices:local"]["Branding:ProductName1"] == "PLENTICORE plus"
        )

    @pytest.mark.asyncio
    async def test_smoketest_setting_value4(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Retrieves the setting value with variante 4."""

        setting_value = await authenticated_client.get_setting_values(
            {"devices:local": ["Branding:ProductName1"]}
        )

        assert setting_value == {
            "devices:local": {"Branding:ProductName1": "PLENTICORE plus"}
        }

    @pytest.mark.asyncio
    async def test_smoketest_process_data_value1(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Retrieves process data values by using str, str variant."""
        process_data = await authenticated_client.get_process_data_values(
            "devices:local", "EM_State"
        )

        assert process_data.keys() == {"devices:local"}
        assert len(process_data["devices:local"]) == 1
        assert process_data["devices:local"]["EM_State"] is not None

    @pytest.mark.asyncio
    async def test_smoketest_process_data_value2(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Retrieves process data values by using str, Iterable[str] variant."""
        process_data = await authenticated_client.get_process_data_values(
            "devices:local", ["EM_State", "Inverter:State"]
        )

        assert process_data.keys() == {"devices:local"}
        assert len(process_data["devices:local"]) == 2
        assert process_data["devices:local"]["EM_State"] is not None
        assert process_data["devices:local"]["Inverter:State"] is not None

    @pytest.mark.asyncio
    async def test_smoketest_process_data_value3(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Retrieves process data values by using Dict[str, Iterable[str]] variant."""
        process_data = await authenticated_client.get_process_data_values(
            {
                "devices:local": ["EM_State", "Inverter:State"],
                "scb:export": ["PortalConActive"],
            }
        )

        assert process_data.keys() == {"devices:local", "scb:export"}
        assert len(process_data["devices:local"]) == 2
        assert process_data["devices:local"]["EM_State"] is not None
        assert process_data["devices:local"]["Inverter:State"] is not None
        assert len(process_data["scb:export"]) == 1
        assert process_data["scb:export"]["PortalConActive"] is not None

    @pytest.mark.asyncio
    async def test_smoketest_read_all_process_values(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Try to read all process values and ensure no exception is thrown."""

        process_data = await authenticated_client.get_process_data()

        for module_id, processdata_ids in process_data.items():
            processdata_values = await authenticated_client.get_process_data_values(
                module_id, processdata_ids
            )

            assert len(processdata_values) == 1
            assert module_id in processdata_values
            assert set(processdata_ids) == set(processdata_values[module_id])
            assert all(
                isinstance(x.unit, str) for x in processdata_values[module_id].values()
            )
            assert all(
                isinstance(x.value, float)
                for x in processdata_values[module_id].values()
            )

    @pytest.mark.asyncio
    async def test_smoketest_read_events(
        self, authenticated_client: pykoplenti.ApiClient
    ):
        """Try to read events from the inverter."""

        events = await authenticated_client.get_events()

        for event in events:
            assert event.start_time < event.end_time



from typing import Any, Callable, Iterable, Union
from unittest.mock import ANY, MagicMock, call

import pytest

import pykoplenti


class _IterableMatcher:
    """Matcher for iterable which does not check the order."""

    def __init__(self, expected: Iterable):
        self._expected = list(expected)

    def __eq__(self, other: object) -> bool:
        if not hasattr(other, "__iter__"):
            return False

        # check if every item in expected matched an item in other
        expected = self._expected.copy()
        for item in other:
            if (idx := expected.index(item)) >= 0:
                del expected[idx]
            else:
                return False

        return len(expected) == 0

    def __str__(self) -> str:
        return str(self._expected)

    def __repr__(self) -> str:
        return repr(self._expected)


class TestVirtualProcessDataValuesDcSum:
    """This class contains tests for virtual process data values for DC sum."""

    @pytest.mark.asyncio
    async def test_virtual_process_data(
        self,
        pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
        client_response_factory: Callable[
            [int, Union[list[Any], dict[Any, Any]]], MagicMock
        ],
        websession: MagicMock,
    ):
        """Test virtual process data for PV power if depencies are present."""
        client_response_factory(
            200,
            [
                {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
                {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
            ],
        )

        values = await pykoplenti_extended_client.get_process_data()

        websession.request.assert_called_once_with(
            "GET",
            ANY,
            headers=ANY,
        )

        assert values == {
            "_virt_": ["pv_P"],
            "devices:local:pv1": ["P"],
            "devices:local:pv2": ["P"],
        }

    @pytest.mark.asyncio
    async def test_virtual_process_data_value(
        self,
        pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
        client_response_factory: Callable[
            [int, Union[list[Any], dict[Any, Any]]], MagicMock
        ],
        websession: MagicMock,
    ):
        """Test virtual process data for PV power."""
        client_response_factory(
            200,
            [
                {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
                {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
                {"moduleid": "devices:local:pv3", "processdataids": ["P"]},
            ],
        )
        client_response_factory(
            200,
            [
                {
                    "moduleid": "devices:local:pv1",
                    "processdata": [
                        {"id": "P", "unit": "W", "value": 700.0},
                    ],
                },
                {
                    "moduleid": "devices:local:pv2",
                    "processdata": [
                        {"id": "P", "unit": "W", "value": 300.0},
                    ],
                },
                {
                    "moduleid": "devices:local:pv3",
                    "processdata": [
                        {"id": "P", "unit": "W", "value": 500.0},
                    ],
                },
            ],
        )

        values = await pykoplenti_extended_client.get_process_data_values(
            "_virt_", "pv_P"
        )

        websession.request.assert_has_calls(
            [
                call("GET", ANY, headers=ANY),
                call(
                    "POST",
                    ANY,
                    headers=ANY,
                    json=[
                        {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
                        {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
                        {"moduleid": "devices:local:pv3", "processdataids": ["P"]},
                    ],
                ),
            ],
            any_order=True,
        )

        assert values == {
            "_virt_": pykoplenti.ProcessDataCollection(
                [pykoplenti.ProcessData(id="pv_P", unit="W", value=1500.0)]
            )
        }


class TestVirtualProcessDataValuesEnergyToGrid:
    """This class contains tests for virtual process data values for energy to grid."""

    @pytest.mark.parametrize("scope", ["Total", "Year", "Month", "Day"])
    @pytest.mark.asyncio
    async def test_virtual_process_data(
        self,
        pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
        client_response_factory: Callable[
            [int, Union[list[Any], dict[Any, Any]]], MagicMock
        ],
        websession: MagicMock,
        scope: str,
    ):
        """Test virtual process data."""
        client_response_factory(
            200,
            [
                {
                    "moduleid": "scb:statistic:EnergyFlow",
                    "processdataids": [
                        f"Statistic:Yield:{scope}",
                        f"Statistic:EnergyHomeBat:{scope}",
                        f"Statistic:EnergyHomePv:{scope}",
                    ],
                },
            ],
        )

        values = await pykoplenti_extended_client.get_process_data()

        websession.request.assert_called_once_with(
            "GET",
            ANY,
            headers=ANY,
        )

        assert values == {
            "_virt_": [f"Statistic:EnergyGrid:{scope}"],
            "scb:statistic:EnergyFlow": [
                f"Statistic:Yield:{scope}",
                f"Statistic:EnergyHomeBat:{scope}",
                f"Statistic:EnergyHomePv:{scope}",
            ],
        }

    @pytest.mark.parametrize("scope", ["Total", "Year", "Month", "Day"])
    @pytest.mark.asyncio
    async def test_virtual_process_data_value(
        self,
        pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
        client_response_factory: Callable[
            [int, Union[list[Any], dict[Any, Any]]], MagicMock
        ],
        websession: MagicMock,
        scope: str,
    ):
        """Test virtuel process data for energy to grid."""
        client_response_factory(
            200,
            [
                {
                    "moduleid": "scb:statistic:EnergyFlow",
                    "processdataids": [
                        f"Statistic:Yield:{scope}",
                        f"Statistic:EnergyHomeBat:{scope}",
                        f"Statistic:EnergyHomePv:{scope}",
                    ],
                },
            ],
        )
        client_response_factory(
            200,
            [
                {
                    "moduleid": "scb:statistic:EnergyFlow",
                    "processdata": [
                        {
                            "id": f"Statistic:Yield:{scope}",
                            "unit": "Wh",
                            "value": 1000.0,
                        },
                        {
                            "id": f"Statistic:EnergyHomeBat:{scope}",
                            "unit": "Wh",
                            "value": 100.0,
                        },
                        {
                            "id": f"Statistic:EnergyHomePv:{scope}",
                            "unit": "Wh",
                            "value": 200.0,
                        },
                    ],
                },
            ],
        )

        values = await pykoplenti_extended_client.get_process_data_values(
            "_virt_", f"Statistic:EnergyGrid:{scope}"
        )

        websession.request.assert_has_calls(
            [
                call("GET", ANY, headers=ANY),
                call(
                    "POST",
                    ANY,
                    headers=ANY,
                    json=[
                        {
                            "moduleid": "scb:statistic:EnergyFlow",
                            "processdataids": _IterableMatcher(
                                {
                                    f"Statistic:Yield:{scope}",
                                    f"Statistic:EnergyHomeBat:{scope}",
                                    f"Statistic:EnergyHomePv:{scope}",
                                }
                            ),
                        },
                    ],
                ),
            ],
            any_order=True,
        )

        assert values == {
            "_virt_": pykoplenti.ProcessDataCollection(
                [
                    pykoplenti.ProcessData(
                        id=f"Statistic:EnergyGrid:{scope}", unit="Wh", value=700.0
                    )
                ]
            )
        }


@pytest.mark.asyncio
async def test_virtual_process_data_no_dc_sum(
    pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
    client_response_factory: Callable[
        [int, Union[list[Any], dict[Any, Any]]], MagicMock
    ],
    websession: MagicMock,
):
    """Test if no virtual process data is present if dependencies are missing."""
    client_response_factory(
        200,
        [
            {"moduleid": "devices:local", "processdataids": ["EM_State"]},
        ],
    )

    values = await pykoplenti_extended_client.get_process_data()

    websession.request.assert_called_once_with(
        "GET",
        ANY,
        headers=ANY,
    )

    assert values == {
        "_virt_": [],
        "devices:local": ["EM_State"],
    }


@pytest.mark.asyncio
async def test_virtual_process_data_and_normal_process_data(
    pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
    client_response_factory: Callable[
        [int, Union[list[Any], dict[Any, Any]]], MagicMock
    ],
    websession: MagicMock,
):
    """Test if virtual and non-virtual process values can be requested."""
    client_response_factory(
        200,
        [
            {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
            {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
        ],
    )
    client_response_factory(
        200,
        [
            {
                "moduleid": "devices:local:pv1",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 700.0},
                ],
            },
            {
                "moduleid": "devices:local:pv2",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 300.0},
                ],
            },
        ],
    )

    values = await pykoplenti_extended_client.get_process_data_values(
        {"_virt_": ["pv_P"], "devices:local:pv1": ["P"], "devices:local:pv2": ["P"]}
    )

    websession.request.assert_has_calls(
        [
            call("GET", ANY, headers=ANY),
            call(
                "POST",
                ANY,
                headers=ANY,
                json=[
                    {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
                    {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
                ],
            ),
        ],
        any_order=True,
    )

    assert values == {
        "_virt_": pykoplenti.ProcessDataCollection(
            [pykoplenti.ProcessData(id="pv_P", unit="W", value=1000.0)]
        ),
        "devices:local:pv1": pykoplenti.ProcessDataCollection(
            [pykoplenti.ProcessData(id="P", unit="W", value=700.0)]
        ),
        "devices:local:pv2": pykoplenti.ProcessDataCollection(
            [pykoplenti.ProcessData(id="P", unit="W", value=300.0)]
        ),
    }


@pytest.mark.asyncio
async def test_virtual_process_data_not_all_requested(
    pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
    client_response_factory: Callable[
        [int, Union[list[Any], dict[Any, Any]]], MagicMock
    ],
    websession: MagicMock,
):
    """Test if not all available virtual process data are requested."""
    client_response_factory(
        200,
        [
            {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
            {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
            {
                "moduleid": "scb:statistic:EnergyFlow",
                "processdataids": [
                    "Statistic:Yield:Total",
                    "Statistic:EnergyHomeBat:Total",
                    "Statistic:EnergyHomePv:Total",
                ],
            },
        ],
    )
    client_response_factory(
        200,
        [
            {
                "moduleid": "devices:local:pv1",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 700.0},
                ],
            },
            {
                "moduleid": "devices:local:pv2",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 300.0},
                ],
            },
        ],
    )

    values = await pykoplenti_extended_client.get_process_data_values(
        {"_virt_": ["pv_P"]}
    )

    websession.request.assert_has_calls(
        [
            call("GET", ANY, headers=ANY),
            call(
                "POST",
                ANY,
                headers=ANY,
                json=[
                    {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
                    {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
                ],
            ),
        ],
        any_order=True,
    )

    assert values == {
        "_virt_": pykoplenti.ProcessDataCollection(
            [pykoplenti.ProcessData(id="pv_P", unit="W", value=1000.0)]
        ),
    }


@pytest.mark.asyncio
async def test_virtual_process_data_multiple_requested(
    pykoplenti_extended_client: pykoplenti.ExtendedApiClient,
    client_response_factory: Callable[
        [int, Union[list[Any], dict[Any, Any]]], MagicMock
    ],
    websession: MagicMock,
):
    """Test if multiple virtual process data are requested."""
    client_response_factory(
        200,
        [
            {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
            {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
            {
                "moduleid": "scb:statistic:EnergyFlow",
                "processdataids": [
                    "Statistic:Yield:Total",
                    "Statistic:EnergyHomeBat:Total",
                    "Statistic:EnergyHomePv:Total",
                ],
            },
        ],
    )
    client_response_factory(
        200,
        [
            {
                "moduleid": "devices:local:pv1",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 700.0},
                ],
            },
            {
                "moduleid": "devices:local:pv2",
                "processdata": [
                    {"id": "P", "unit": "W", "value": 300.0},
                ],
            },
            {
                "moduleid": "scb:statistic:EnergyFlow",
                "processdata": [
                    {
                        "id": "Statistic:Yield:Total",
                        "unit": "Wh",
                        "value": 1000.0,
                    },
                    {
                        "id": "Statistic:EnergyHomeBat:Total",
                        "unit": "Wh",
                        "value": 100.0,
                    },
                    {
                        "id": "Statistic:EnergyHomePv:Total",
                        "unit": "Wh",
                        "value": 200.0,
                    },
                ],
            },
        ],
    )

    values = await pykoplenti_extended_client.get_process_data_values(
        {"_virt_": ["pv_P", "Statistic:EnergyGrid:Total"]}
    )

    websession.request.assert_has_calls(
        [
            call("GET", ANY, headers=ANY),
            call(
                "POST",
                ANY,
                headers=ANY,
                json=_IterableMatcher(
                    [
                        {"moduleid": "devices:local:pv1", "processdataids": ["P"]},
                        {"moduleid": "devices:local:pv2", "processdataids": ["P"]},
                        {
                            "moduleid": "scb:statistic:EnergyFlow",
                            "processdataids": _IterableMatcher(
                                {
                                    "Statistic:Yield:Total",
                                    "Statistic:EnergyHomeBat:Total",
                                    "Statistic:EnergyHomePv:Total",
                                }
                            ),
                        },
                    ]
                ),
            ),
        ],
        any_order=True,
    )

    assert values == {
        "_virt_": pykoplenti.ProcessDataCollection(
            [
                pykoplenti.ProcessData(id="pv_P", unit="W", value=1000.0),
                pykoplenti.ProcessData(
                    id="Statistic:EnergyGrid:Total", unit="Wh", value=700.0
                ),
            ]
        ),
    }



from ast import literal_eval
import asyncio
from collections import defaultdict
from inspect import iscoroutinefunction
import os
from pprint import pprint
import re
import tempfile
import traceback
from typing import Any, Awaitable, Callable, Dict, Union

from aiohttp import ClientSession, ClientTimeout
import click
from prompt_toolkit import PromptSession, print_formatted_text

from pykoplenti import ApiClient
from pykoplenti.extended import ExtendedApiClient


class SessionCache:
    """Persistent the session in a temporary file."""

    def __init__(self, host):
        self.host = host

    def read_session_id(self) -> Union[str, None]:
        file = os.path.join(tempfile.gettempdir(), f"pykoplenti-session-{self.host}")
        if os.path.isfile(file):
            with open(file, "rt") as f:
                return f.readline(256)
        else:
            return None

    def write_session_id(self, id: str):
        file = os.path.join(tempfile.gettempdir(), f"pykoplenti-session-{self.host}")
        f = os.open(file, os.O_WRONLY | os.O_TRUNC | os.O_CREAT, mode=0o600)
        try:
            os.write(f, id.encode("ascii"))
        finally:
            os.close(f)


class ApiShell:
    """Provides a shell-like access to the inverter."""

    def __init__(self, client: ApiClient):
        super().__init__()
        self.client = client
        self._session_cache = SessionCache(self.client.host)

    async def prepare_client(self, passwd):
        # first try to reuse existing session
        session_id = self._session_cache.read_session_id()
        if session_id is not None:
            self.client.session_id = session_id
            print_formatted_text("Trying to reuse existing session... ", end=None)
            me = await self.client.get_me()
            if me.is_authenticated:
                print_formatted_text("Success")
                return

            print_formatted_text("Failed")

        if passwd is not None:
            print_formatted_text("Logging in... ", end=None)
            await self.client.login(passwd)
            self._session_cache.write_session_id(self.client.session_id)
            print_formatted_text("Success")

    def print_exception(self):
        """Prints an excpetion from executing a method."""
        print_formatted_text(traceback.format_exc())

    async def run(self, passwd):
        session = PromptSession()
        print_formatted_text(flush=True)  # Initialize output

        # Test commands:
        # get_settings
        # get_setting_values 'devices:local' 'Battery:MinSoc'
        # get_setting_values 'devices:local' ['Battery:MinSoc', \
        # 'Battery:MinHomeComsumption']
        # get_setting_values 'scb:time'
        # set_setting_values 'devices:local' {'Battery:MinSoc':'15'}

        await self.prepare_client(passwd)

        while True:
            try:
                text = await session.prompt_async("(pykoplenti)> ")

                if text.strip().lower() == "exit":
                    raise EOFError()

                if text.strip() == "":
                    continue
                else:
                    # TODO split does not know about lists or dicts or strings
                    # with spaces
                    method_name, *arg_values = text.strip().split()

                    if method_name == "help":
                        self._do_help(arg_values)
                        continue

                    method = self._get_method(method_name)
                    if method is None:
                        continue

                    args = self._create_args(arg_values)
                    if args is None:
                        continue

                    await self._execute(method, args)

            except KeyboardInterrupt:
                continue
            except EOFError:
                break

    def _do_help(self, argv):
        if len(argv) == 0:
            print_formatted_text("Try: help <command>")
        else:
            method = getattr(self.client, argv[0])
            print_formatted_text(method.__doc__)

    def _get_method(self, name):
        try:
            return getattr(self.client, name)
        except AttributeError:
            print_formatted_text(f"Unknown method: {name}")
            return None

    def _create_args(self, argv):
        try:
            return [literal_eval(x) for x in argv]
        except Exception:
            print_formatted_text("Error parsing arguments")
            self.print_exception()
            return None

    async def _execute(self, method, args):
        try:
            if iscoroutinefunction(method):
                result = await method(*args)
            else:
                result = method(*args)

            pprint(result)
        except Exception:
            print_formatted_text("Error executing method")
            self.print_exception()


async def repl_main(host, port, passwd):
    async with ClientSession(timeout=ClientTimeout(total=10)) as session:
        client = ExtendedApiClient(session, host=host, port=port)

        shell = ApiShell(client)
        await shell.run(passwd)


async def command_main(
    host: str, port: int, passwd: str, fn: Callable[[ApiClient], Awaitable[Any]]
):
    async with ClientSession(timeout=ClientTimeout(total=10)) as session:
        client = ExtendedApiClient(session, host=host, port=port)
        session_cache = SessionCache(host)

        # Try to reuse an existing session
        client.session_id = session_cache.read_session_id()
        me = await client.get_me()
        if not me.is_authenticated:
            # create a new session
            await client.login(passwd)
            if client.session_id is not None:
                session_cache.write_session_id(client.session_id)

        await fn(client)


class GlobalArgs:
    """Global arguments over all sub commands."""

    def __init__(self):
        self.host = None
        self.port = None
        self.password = None
        self.password_file = None


pass_global_args = click.make_pass_decorator(GlobalArgs, ensure=True)


@click.group()
@click.option("--host", help="hostname or ip of the inverter")
@click.option("--port", default=80, help="port of the inverter (default 80)")
@click.option("--password", default=None, help="the password")
@click.option(
    "--password-file",
    default="secrets",
    help='password file (default "secrets" in the current working directory)',
)
@pass_global_args
def cli(global_args, host, port, password, password_file):
    """Handling of global arguments with click"""
    if password is not None:
        global_args.passwd = password
    elif os.path.isfile(password_file):
        with open(password_file, "rt") as f:
            global_args.passwd = f.readline()
    else:
        global_args.passwd = None

    global_args.host = host
    global_args.port = port


@cli.command()
@pass_global_args
def repl(global_args):
    """Provides a simple REPL for executing API requests to the inverter."""
    asyncio.run(repl_main(global_args.host, global_args.port, global_args.passwd))


@cli.command()
@click.option("--lang", default=None, help="language for events")
@click.option("--count", default=10, help="number of events to read")
@pass_global_args
def read_events(global_args, lang, count):
    """Returns the last events"""

    async def fn(client: ApiClient):
        data = await client.get_events(lang=lang, max_count=count)
        for event in data:
            print(
                f"{event.is_active < 5} {event.start_time} {event.end_time} "
                f"{event.description}"
            )

    asyncio.run(
        command_main(global_args.host, global_args.port, global_args.passwd, fn)
    )


@cli.command()
@click.option(
    "--out",
    required=True,
    type=click.File(mode="wt", encoding="UTF-8"),
    help="file to write the log data to",
)
@click.option("--begin", type=click.DateTime(["%Y-%m-%d"]), help="first day to export")
@click.option("--end", type=click.DateTime(["%Y-%m-%d"]), help="last day to export")
@pass_global_args
def download_log(global_args, out, begin, end):
    """Download the log data from the inverter to a file."""

    async def fn(client: ApiClient):
        await client.download_logdata(writer=out, begin=begin, end=end)

    asyncio.run(
        command_main(global_args.host, global_args.port, global_args.passwd, fn)
    )


@cli.command()
@pass_global_args
def all_processdata(global_args):
    """Returns a list of all available process data."""

    async def fn(client: ApiClient):
        data = await client.get_process_data()
        for k, v in data.items():
            for x in v:
                print(f"{k}/{x}")

    asyncio.run(
        command_main(global_args.host, global_args.port, global_args.passwd, fn)
    )


@cli.command()
@click.argument("ids", required=True, nargs=-1)
@pass_global_args
def read_processdata(global_args, ids):
    """Returns the values of the given process data.

    IDS is the identifier (<module_id>/<processdata_id>) of one or more processdata
    to read.

    \b
    Examples:
        read-processdata devices:local/Inverter:State
    """

    async def fn(client: ApiClient):
        if len(ids) == 1 and "/" not in ids[0]:
            # all process data ids of a moudle
            values = await client.get_process_data_values(ids[0])
        else:
            query = defaultdict(list)
            for id in ids:
                m = re.match(r"(?P<module_id>.+)/(?P<processdata_id>.+)", id)
                if not m:
                    raise Exception(f"Invalid format of {id}")

                module_id = m.group("module_id")
                setting_id = m.group("processdata_id")

                query[module_id].append(setting_id)

            values = await client.get_process_data_values(query)

        for k, v in values.items():
            for x in v.values():
                print(f"{k}/{x.id}={x.value}")

    asyncio.run(
        command_main(global_args.host, global_args.port, global_args.passwd, fn)
    )


@cli.command()
@click.option(
    "--rw", is_flag=True, default=False, help="display only writable settings"
)
@pass_global_args
def all_settings(global_args, rw):
    """Returns the ids of all settings."""

    async def fn(client: ApiClient):
        settings = await client.get_settings()
        for k, v in settings.items():
            for x in v:
                if not rw or x.access == "readwrite":
                    print(f"{k}/{x.id}")

    asyncio.run(
        command_main(global_args.host, global_args.port, global_args.passwd, fn)
    )


@cli.command()
@click.argument("ids", required=True, nargs=-1)
@pass_global_args
def read_settings(global_args, ids):
    """Read the value of the given settings.

    IDS is the identifier (<module_id>/<setting_id>) of one or more settings to read

    \b
    Examples:
        read-settings devices:local/Battery:MinSoc
        read-settings devices:local/Battery:MinSoc \
            devices:local/Battery:MinHomeComsumption
    """

    async def fn(client: ApiClient):
        query = defaultdict(list)
        for id in ids:
            m = re.match(r"(?P<module_id>.+)/(?P<setting_id>.+)", id)
            if not m:
                raise Exception(f"Invalid format of {id}")

            module_id = m.group("module_id")
            setting_id = m.group("setting_id")

            query[module_id].append(setting_id)

        values = await client.get_setting_values(query)

        for k, x in values.items():
            for i, v in x.items():
                print(f"{k}/{i}={v}")

    asyncio.run(
        command_main(global_args.host, global_args.port, global_args.passwd, fn)
    )


@cli.command()
@click.argument("id_values", required=True, nargs=-1)
@pass_global_args
def write_settings(global_args, id_values):
    """Write the values of the given settings.

    ID_VALUES is the identifier plus the the value to write

    \b
    Examples:
        write-settings devices:local/Battery:MinSoc=15
    """

    async def fn(client: ApiClient):
        query: Dict[str, Dict[str, str]] = defaultdict(dict)
        for id_value in id_values:
            m = re.match(
                r"(?P<module_id>.+)/(?P<setting_id>.+)=(?P<value>.+)", id_value
            )
            if not m:
                raise Exception(f"Invalid format of {id_value}")

            module_id = m.group("module_id")
            setting_id = m.group("setting_id")
            value = m.group("value")

            query[module_id][setting_id] = value

        for module_id, setting_values in query.items():
            await client.set_setting_values(module_id, setting_values)

    asyncio.run(
        command_main(global_args.host, global_args.port, global_args.passwd, fn)
    )


# entry point for pycharm; should not be used for commandline usage
if __name__ == "__main__":
    import sys

    cli(sys.argv[1:], auto_envvar_prefix="PYKOPLENTI")



"""Extended ApiClient which provides virtual process data values."""

from abc import ABC, abstractmethod
from collections import ChainMap, defaultdict
from typing import Final, Iterable, Literal, Mapping, MutableMapping, Union

from aiohttp import ClientSession

from .api import ApiClient
from .model import ProcessData, ProcessDataCollection

_VIRT_MODUL_ID: Final = "_virt_"


class _VirtProcessDataItemBase(ABC):
    """Base class for all virtual process data items."""

    def __init__(self, processid: str, process_data: dict[str, set[str]]) -> None:
        self.processid = processid
        self.process_data = process_data
        self.available_process_data: dict[str, set[str]] = {}

    def update_actual_process_ids(
        self, available_process_ids: Mapping[str, Iterable[str]]
    ):
        """Update which process data for this item are available."""
        self.available_process_data.clear()
        for module_id, process_ids in self.process_data.items():
            if module_id in available_process_ids:
                matching_process_ids = process_ids.intersection(
                    available_process_ids[module_id]
                )
                if len(matching_process_ids) > 0:
                    self.available_process_data[module_id] = matching_process_ids

    @abstractmethod
    def get_value(
        self, process_values: Mapping[str, ProcessDataCollection]
    ) -> ProcessData:
        ...

    @abstractmethod
    def is_available(self) -> bool:
        ...


class _VirtProcessDataItemSum(_VirtProcessDataItemBase):
    def get_value(
        self, process_values: Mapping[str, ProcessDataCollection]
    ) -> ProcessData:
        values: list[float] = []
        for module_id, process_ids in self.available_process_data.items():
            values += (process_values[module_id][pid].value for pid in process_ids)

        return ProcessData(id=self.processid, unit="W", value=sum(values))

    def is_available(self) -> bool:
        return len(self.available_process_data) > 0


class _VirtProcessDataItemEnergyToGrid(_VirtProcessDataItemBase):
    def __init__(
        self, processid: str, scope: Literal["Total", "Year", "Month", "Day"]
    ) -> None:
        super().__init__(
            processid,
            {
                "scb:statistic:EnergyFlow": {
                    f"Statistic:Yield:{scope}",
                    f"Statistic:EnergyHomeBat:{scope}",
                    f"Statistic:EnergyHomePv:{scope}",
                }
            },
        )
        self.scope = scope

    def get_value(
        self, process_values: Mapping[str, ProcessDataCollection]
    ) -> ProcessData:
        statistics = process_values["scb:statistic:EnergyFlow"]
        energy_yield = statistics[f"Statistic:Yield:{self.scope}"].value
        energy_home_bat = statistics[f"Statistic:EnergyHomeBat:{self.scope}"].value
        energy_home_pv = statistics[f"Statistic:EnergyHomePv:{self.scope}"].value

        return ProcessData(
            id=self.processid,
            unit="Wh",
            value=energy_yield - energy_home_pv - energy_home_bat,
        )

    def is_available(self) -> bool:
        return self.available_process_data == self.process_data


class _VirtProcessDataManager:
    """Manager for all virtual process data items."""

    def __init__(self) -> None:
        self._virt_process_data: Iterable[_VirtProcessDataItemBase] = [
            _VirtProcessDataItemSum(
                "pv_P",
                {
                    "devices:local:pv1": {"P"},
                    "devices:local:pv2": {"P"},
                    "devices:local:pv3": {"P"},
                },
            ),
            _VirtProcessDataItemEnergyToGrid("Statistic:EnergyGrid:Total", "Total"),
            _VirtProcessDataItemEnergyToGrid("Statistic:EnergyGrid:Year", "Year"),
            _VirtProcessDataItemEnergyToGrid("Statistic:EnergyGrid:Month", "Month"),
            _VirtProcessDataItemEnergyToGrid("Statistic:EnergyGrid:Day", "Day"),
        ]

    def initialize(self, available_process_data: Mapping[str, Iterable[str]]):
        """Initialize the virtual items with the list of available process ids."""
        for vpd in self._virt_process_data:
            vpd.update_actual_process_ids(available_process_data)

    def adapt_process_data_response(
        self, process_data: dict[str, list[str]]
    ) -> Mapping[str, list[str]]:
        """Adapt the reponse of reading process data."""
        virt_process_data: dict[str, list[str]] = {_VIRT_MODUL_ID: []}

        for vpd in self._virt_process_data:
            if vpd.is_available():
                virt_process_data[_VIRT_MODUL_ID].append(vpd.processid)

        return ChainMap(process_data, virt_process_data)

    def adapt_process_value_request(
        self, process_data: Mapping[str, Iterable[str]]
    ) -> Mapping[str, Iterable[str]]:
        """Adapt the request for process values."""
        result: MutableMapping[str, set[str]] = defaultdict(set)

        for mid, pids in process_data.items():
            result[mid].update(pids)

        for requested_virtual_process_id in result.pop(_VIRT_MODUL_ID):
            for virtual_process_data in self._virt_process_data:
                if virtual_process_data.is_available():
                    if requested_virtual_process_id == virtual_process_data.processid:
                        # add ids for virtual if they are missing
                        for (
                            mid,
                            pids,
                        ) in virtual_process_data.available_process_data.items():
                            result[mid].update(pids)
                        break
            else:
                raise ValueError(
                    f"No virtual process data '{requested_virtual_process_id}'."
                )

        return result

    def adapt_process_value_response(
        self,
        values: Mapping[str, ProcessDataCollection],
        request_data: Mapping[str, Iterable[str]],
    ) -> Mapping[str, ProcessDataCollection]:
        """Adapt the reponse for process values."""
        result = {}

        # add virtual items
        virtual_process_data_values = []
        for id in request_data[_VIRT_MODUL_ID]:
            for vpd in self._virt_process_data:
                if vpd.processid == id:
                    virtual_process_data_values.append(vpd.get_value(values))
        result["_virt_"] = ProcessDataCollection(virtual_process_data_values)

        # add all values which was requested but not the extra ids for the virtual ids
        for mid, pdc in values.items():
            if mid in request_data:
                pids = [x for x in pdc.values() if x.id in request_data[mid]]
                if len(pids) > 0:
                    result[mid] = ProcessDataCollection(pids)

        return result


class ExtendedApiClient(ApiClient):
    """Extend ApiClient with virtual process data."""

    def __init__(self, websession: ClientSession, host: str, port: int = 80):
        super().__init__(websession, host, port)

        self._virt_process_data = _VirtProcessDataManager()
        self._virt_process_data_initialized = False

    async def get_process_data(self) -> Mapping[str, Iterable[str]]:
        process_data = await super().get_process_data()

        self._virt_process_data.initialize(process_data)
        self._virt_process_data_initialized = True
        return self._virt_process_data.adapt_process_data_response(process_data)

    async def get_process_data_values(
        self,
        module_id: Union[str, Mapping[str, Iterable[str]]],
        processdata_id: Union[str, Iterable[str], None] = None,
    ) -> Mapping[str, ProcessDataCollection]:
        contains_virt_process_data = (
            isinstance(module_id, str) and _VIRT_MODUL_ID == module_id
        ) or (isinstance(module_id, dict) and _VIRT_MODUL_ID in module_id)

        if not contains_virt_process_data:
            # short-cut if no virtual process is requested
            return await super().get_process_data_values(module_id, processdata_id)

        process_data: dict[str, Iterable[str]] = {}
        if isinstance(module_id, str) and processdata_id is None:
            process_data[module_id] = []
        elif isinstance(module_id, str) and isinstance(processdata_id, str):
            process_data[module_id] = [processdata_id]
        elif (
            isinstance(module_id, str)
            and processdata_id is not None
            and hasattr(processdata_id, "__iter__")
        ):
            process_data[module_id] = list(processdata_id)
        elif isinstance(module_id, Mapping) and processdata_id is None:
            process_data.update(module_id)
        else:
            raise TypeError("Invalid combination of module_id and processdata_id.")

        if not self._virt_process_data_initialized:
            pd = await self.get_process_data()
            self._virt_process_data.initialize(pd)
            self._virt_process_data_initialized = True

        process_values = await super().get_process_data_values(
            self._virt_process_data.adapt_process_value_request(process_data)
        )
        return self._virt_process_data.adapt_process_value_response(
            process_values, process_data
        )



from datetime import datetime
from typing import Final, Iterator, Mapping, Optional

import pydantic
from pydantic import BaseModel, Field


class MeData(BaseModel):
    """Represent the data of the 'me'-request."""

    is_locked: bool = Field(alias="locked")
    is_active: bool = Field(alias="active")
    is_authenticated: bool = Field(alias="authenticated")
    permissions: list[str] = Field()
    is_anonymous: bool = Field(alias="anonymous")
    role: str


class VersionData(BaseModel):
    """Represent the data of the 'version'-request."""

    api_version: str
    hostname: str
    name: str
    sw_version: str


class ModuleData(BaseModel):
    """Represents a single module."""

    id: str
    type: str


class ProcessData(BaseModel):
    """Represents a single process data."""

    id: str
    unit: str
    value: float


class ProcessDataCollection(Mapping):
    """Represents a collection of process data value."""

    def __init__(self, process_data: list[ProcessData]):
        self._process_data = process_data

    def __len__(self) -> int:
        return len(self._process_data)

    def __iter__(self) -> Iterator[str]:
        return (x.id for x in self._process_data)

    def __getitem__(self, item) -> ProcessData:
        try:
            return next(x for x in self._process_data if x.id == item)
        except StopIteration:
            raise KeyError(item) from None

    def __eq__(self, __other: object) -> bool:
        if not isinstance(__other, ProcessDataCollection):
            return False

        return self._process_data == __other._process_data

    def __str__(self):
        return "[" + ",".join(str(x) for x in self._process_data) + "]"

    def __repr__(self):
        return (
            "ProcessDataCollection(["
            + ",".join(repr(x) for x in self._process_data)
            + "])"
        )


class SettingsData(BaseModel):
    """Represents a single settings data."""

    min: Optional[str]
    max: Optional[str]
    default: Optional[str]
    access: str
    unit: Optional[str]
    id: str
    type: str


class EventData(BaseModel):
    """Represents an event of the inverter."""

    start_time: datetime
    end_time: datetime
    code: int
    long_description: str
    category: str
    description: str
    group: str
    is_active: bool


# pydantic version specific code
# In pydantic 2.x `parse_obj_as` is no longer supported. To stay compatible to
# both version a small wrapper function is used.

if pydantic.VERSION.startswith("2."):
    from pydantic import TypeAdapter

    _process_list_adapter: Final = TypeAdapter(list[ProcessData])

    def process_data_list(json) -> list[ProcessData]:
        """Process json as a list of ProcessData objects."""
        return _process_list_adapter.validate_python(json)

else:
    from pydantic import parse_obj_as

    def process_data_list(json) -> list[ProcessData]:
        """Process json as a list of ProcessData objects."""
        return parse_obj_as(list[ProcessData], json)
